from __future__ import annotations

import asyncio
import inspect
import sys
from collections.abc import Iterator  # pylint: disable=import-error
import random
from typing import List, TypeVar, AsyncIterable, Iterable, AsyncIterator, cast
import string

from sqlglot import expressions as exp
from sqlglot.optimizer.scope import traverse_scope


T = TypeVar("T")

# MySQL Connector/J uses ASCII to decode nonce
SAFE_NONCE_CHARS = (string.ascii_letters + string.digits).encode()


class seq(Iterator):
    """Auto-incrementing sequence with an optional maximum size"""

    def __init__(self, size: int | None = None):
        self.size = size
        self.value = 0

    def __next__(self) -> int:
        value = self.value
        self.value = self.value + 1
        if self.size:
            self.value = self.value % self.size
        return value

    def reset(self) -> None:
        self.value = 0


def xor(a: bytes, b: bytes) -> bytes:
    # Fast XOR implementation, according to https://stackoverflow.com/questions/29408173/byte-operations-xor-in-python
    a, b = a[: len(b)], b[: len(a)]
    int_b = int.from_bytes(b, sys.byteorder)
    int_a = int.from_bytes(a, sys.byteorder)
    int_enc = int_b ^ int_a
    return int_enc.to_bytes(len(b), sys.byteorder)


def nonce(nbytes: int) -> bytes:
    return bytes(
        bytearray(
            [random.SystemRandom().choice(SAFE_NONCE_CHARS) for _ in range(nbytes)]
        )
    )


def find_tables(expression: exp.Expression) -> List[exp.Table]:
    """Find all tables in an expression"""
    if isinstance(
        expression, (exp.Select, exp.Subquery, exp.Union, exp.Except, exp.Intersect)
    ):
        return [
            source
            for scope in traverse_scope(expression)
            for source in scope.sources.values()
            if isinstance(source, exp.Table)
        ]
    return []


def find_dbs(expression: exp.Expression) -> List[str]:
    """Find all database names in an expression"""
    return [table.text("db") for table in find_tables(expression)]


def dict_depth(d: dict) -> int:
    """
    Get the nesting depth of a dictionary.
    For example:
        >>> dict_depth(None)
        0
        >>> dict_depth({})
        1
        >>> dict_depth({"a": "b"})
        1
        >>> dict_depth({"a": {}})
        2
        >>> dict_depth({"a": {"b": {}}})
        3
    Args:
        d (dict): dictionary
    Returns:
        int: depth
    """
    try:
        return 1 + dict_depth(next(iter(d.values())))
    except AttributeError:
        # d doesn't have attribute "values"
        return 0
    except StopIteration:
        # d.values() returns an empty sequence
        return 1


async def aiterate(iterable: AsyncIterable[T] | Iterable[T]) -> AsyncIterator[T]:
    """Iterate either an async iterable or a regular iterable"""
    if inspect.isasyncgen(iterable):
        async for item in iterable:
            yield item
    else:
        for item in cast(Iterable, iterable):
            yield item


async def cooperative_iterate(
    iterable: AsyncIterable[T], batch_size: int = 10_000
) -> AsyncIterator[T]:
    """
    Iterate an async iterable in a cooperative manner, yielding control back to the event loop every `batch_size` iterations
    """
    i = 0
    async for item in iterable:
        if i != 0 and i % batch_size == 0:
            await asyncio.sleep(0)
        yield item
        i += 1



import asyncio
from contextlib import closing
from typing import Callable

import pytest
from mysql.connector import DatabaseError

from tests.conftest import MockSession, query, ConnectFixture


@pytest.fixture
def session_factory(session: MockSession) -> Callable[[], MockSession]:
    called_once = False

    def factory() -> MockSession:
        nonlocal called_once
        if not called_once:
            called_once = True
            return session
        return MockSession()

    return factory


@pytest.mark.asyncio
async def test_kill_connection(
    session: MockSession,
    connect: ConnectFixture,
) -> None:
    session.pause = asyncio.Event()

    with closing(await connect()) as conn1:
        with closing(await connect()) as conn2:
            q1 = asyncio.create_task(query(conn1, "SELECT a FROM x"))

            # Wait until we know the session is paused
            await session.waiting.wait()

            await query(conn2, f"KILL {session.connection.connection_id}")

            with pytest.raises(DatabaseError) as ctx:
                await q1

            assert "Session was killed" in str(ctx.value.msg)

            session.pause.clear()

            with pytest.raises(DatabaseError) as ctx:
                await query(conn1, "SELECT 1")

            assert "Connection not available" in str(ctx.value.msg)


@pytest.mark.asyncio
async def test_kill_query(
    session: MockSession,
    connect: ConnectFixture,
) -> None:
    session.pause = asyncio.Event()

    with closing(await connect()) as conn1:
        with closing(await connect()) as conn2:
            q1 = asyncio.create_task(query(conn1, "SELECT a FROM x"))

            # Wait until we know the session is paused
            await session.waiting.wait()

            await query(conn2, f"KILL QUERY {session.connection.connection_id}")

            with pytest.raises(DatabaseError) as ctx:
                await q1

            assert "Query was killed" in str(ctx.value.msg)

            session.pause.clear()

            assert (await query(conn1, "SELECT 1")) == [{"1": 1}]



from __future__ import annotations

import asyncio
import inspect
import logging
from ssl import SSLContext
from socket import socket
from typing import Callable, Any, Optional, Sequence, Awaitable

from mysql_mimic import packets
from mysql_mimic.auth import IdentityProvider, SimpleIdentityProvider
from mysql_mimic.connection import Connection
from mysql_mimic.control import Control, LocalControl, TooManyConnections
from mysql_mimic.errors import ErrorCode
from mysql_mimic.session import Session, BaseSession
from mysql_mimic.constants import DEFAULT_SERVER_CAPABILITIES
from mysql_mimic.stream import MysqlStream
from mysql_mimic.types import Capabilities


logger = logging.getLogger(__name__)


class MaxConnectionsExceeded(Exception):
    pass


class MysqlServer:
    """
    MySQL mimic server.

    Args:
        session_factory: Callable that takes no arguments and returns a session
        capabilities: server capability flags
        control: Control instance to use. Defaults to a LocalControl instance.
        identity_provider: Authentication plugins to register. Defaults to `SimpleIdentityProvider`,
            which just blindly accepts whatever `username` is given by the client.
        ssl: SSLContext instance if this server should enable TLS over connections

        **kwargs: extra keyword args passed to the asyncio start server command
    """

    def __init__(
        self,
        session_factory: Callable[[], BaseSession | Awaitable[BaseSession]] = Session,
        capabilities: Capabilities = DEFAULT_SERVER_CAPABILITIES,
        control: Control | None = None,
        identity_provider: IdentityProvider | None = None,
        ssl: SSLContext | None = None,
        **serve_kwargs: Any,
    ):
        self.session_factory = session_factory
        self.capabilities = capabilities
        self.identity_provider = identity_provider or SimpleIdentityProvider()
        self.ssl = ssl

        self.control = control or LocalControl()
        self._serve_kwargs = serve_kwargs
        self._server: Optional[asyncio.base_events.Server] = None

    async def _client_connected_cb(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        stream = MysqlStream(reader, writer)

        try:
            if inspect.iscoroutinefunction(self.session_factory):
                session = await self.session_factory()
            else:
                session = self.session_factory()

            connection = Connection(
                stream=stream,
                session=session,
                control=self.control,
                server_capabilities=self.capabilities,
                identity_provider=self.identity_provider,
                ssl=self.ssl,
            )

        except Exception:  # pylint: disable=broad-except
            logger.exception("Failed to create connection")
            await stream.write(
                # Return an error so clients don't freeze
                packets.make_error(capabilities=self.capabilities)
            )
            return

        try:
            connection_id = await self.control.add(connection)
            connection.connection_id = connection_id
        except TooManyConnections:
            await stream.write(
                connection.error(
                    msg="Too many connections",
                    code=ErrorCode.CON_COUNT_ERROR,
                )
            )
            return
        except Exception:  # pylint: disable=broad-except
            logger.exception("Failed to register connection")
            await stream.write(connection.error(msg="Failed to register connection"))
            return

        try:
            return await connection.start()
        finally:
            writer.close()
            await self.control.remove(connection_id)

    async def start_server(self, **kwargs: Any) -> None:
        """
        Start an asyncio socket server.

        Args:
            **kwargs: keyword args passed to `asyncio.start_server`
        """
        kw = {}
        kw.update(self._serve_kwargs)
        kw.update(kwargs)
        if "port" not in kw:
            kw["port"] = 3306
        self._server = await asyncio.start_server(self._client_connected_cb, **kw)

    async def start_unix_server(self, **kwargs: Any) -> None:
        """
        Start an asyncio unix socket server.

        Args:
            **kwargs: keyword args passed to `asyncio.start_unix_server`
        """
        kw = {}
        kw.update(self._serve_kwargs)
        kw.update(kwargs)
        self._server = await asyncio.start_unix_server(self._client_connected_cb, **kw)

    async def serve_forever(self, **kwargs: Any) -> None:
        """
        Start accepting connections until the coroutine is cancelled.

        If a server isn't running, this starts a server with `start_server`.

        Args:
            **kwargs: keyword args passed to `start_server`
        """
        if not self._server:
            await self.start_server(**kwargs)
        assert self._server is not None
        await self._server.serve_forever()

    def close(self) -> None:
        """
        Stop serving.

        The server is closed asynchronously -
        use the `wait_closed` coroutine to wait until the server is closed.
        """
        if self._server:
            self._server.close()

    async def wait_closed(self) -> None:
        """Wait until the `close` method completes."""
        if self._server:
            await self._server.wait_closed()

    def sockets(self) -> Sequence[socket]:
        """Get sockets the server is listening on."""
        if self._server:
            return self._server.sockets
        return ()



from __future__ import annotations

from enum import auto, Enum

from mysql_mimic.types import Capabilities

DEFAULT_SERVER_CAPABILITIES = (
    Capabilities.CLIENT_PROTOCOL_41
    | Capabilities.CLIENT_DEPRECATE_EOF
    | Capabilities.CLIENT_CONNECT_WITH_DB
    | Capabilities.CLIENT_QUERY_ATTRIBUTES
    | Capabilities.CLIENT_CONNECT_ATTRS
    | Capabilities.CLIENT_PLUGIN_AUTH
    | Capabilities.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA
    | Capabilities.CLIENT_SECURE_CONNECTION
    | Capabilities.CLIENT_LONG_PASSWORD
    | Capabilities.CLIENT_ODBC
    | Capabilities.CLIENT_INTERACTIVE
    | Capabilities.CLIENT_IGNORE_SPACE
)


class KillKind(Enum):
    # Terminate the statement the connection is currently executing, but leave the connection itself intact
    QUERY = auto()
    # Terminate the connection, after terminating any statement the connection is executing
    CONNECTION = auto()


INFO_SCHEMA = {
    "information_schema": {
        "character_sets": {
            "character_set_name": "TEXT",
            "default_collate_name": "TEXT",
            "description": "TEXT",
            "maxlen": "INT",
        },
        "collations": {
            "collation_name": "TEXT",
            "character_set_name": "TEXT",
            "id": "TEXT",
            "is_default": "TEXT",
            "is_compiled": "TEXT",
            "sortlen": "INT",
        },
        "columns": {
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "column_name": "TEXT",
            "ordinal_position": "INT",
            "column_default": "TEXT",
            "is_nullable": "TEXT",
            "data_type": "TEXT",
            "character_maximum_length": "INT",
            "character_octet_length": "INT",
            "numeric_precision": "INT",
            "numeric_scale": "INT",
            "datetime_precision": "INT",
            "character_set_name": "TEXT",
            "collation_name": "TEXT",
            "column_type": "TEXT",
            "column_key": "TEXT",
            "extra": "TEXT",
            "privileges": "TEXT",
            "column_comment": "TEXT",
            "generation_expression": "TEXT",
            "srs_id": "TEXT",
        },
        "column_privileges": {
            "grantee": "TEXT",
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "column_name": "TEXT",
            "privilege_type": "TEXT",
            "is_grantable": "TEXT",
        },
        "events": {
            "event_catalog": "TEXT",
            "event_schema": "TEXT",
            "event_name": "TEXT",
            "definer": "TEXT",
            "time_zone": "TEXT",
            "event_body": "TEXT",
            "event_definition": "TEXT",
            "event_type": "TEXT",
            "execute_at": "TEXT",
            "interval_value": "INT",
            "interval_field": "TEXT",
            "sql_mode": "TEXT",
            "starts": "TEXT",
            "ends": "TEXT",
            "status": "TEXT",
            "on_completion": "TEXT",
            "created": "TEXT",
            "last_altered": "TEXT",
            "last_executed": "TEXT",
            "event_comment": "TEXT",
            "originator": "INT",
            "character_set_client": "TEXT",
            "collation_connection": "TEXT",
            "database_collation": "TEXT",
        },
        "key_column_usage": {
            "constraint_catalog": "TEXT",
            "constraint_schema": "TEXT",
            "constraint_name": "TEXT",
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "column_name": "TEXT",
            "ordinal_position": "INT",
            "position_in_unique_constraint": "INT",
            "referenced_table_schema": "TEXT",
            "referenced_table_name": "TEXT",
            "referenced_column_name": "TEXT",
        },
        "parameters": {
            "specific_catalog": "TEXT",
            "specific_schema": "TEXT",
            "specific_name": "TEXT",
            "ordinal_position": "INT",
            "parameter_mode": "TEXT",
            "parameter_name": "TEXT",
            "data_type": "TEXT",
            "character_maximum_length": "INT",
            "character_octet_length": "INT",
            "numeric_precision": "INT",
            "numeric_scale": "INT",
            "datetime_precision": "INT",
            "character_set_name": "TEXT",
            "collation_name": "TEXT",
            "dtd_identifier": "TEXT",
            "routine_type": "TEXT",
        },
        "partitions": {
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "partition_name": "TEXT",
            "subpartition_name": "TEXT",
            "partition_ordinal_position": "INT",
            "subpartition_ordinal_position": "INT",
            "partition_method": "TEXT",
            "subpartition_method": "TEXT",
            "partition_expression": "TEXT",
            "subpartition_expression": "TEXT",
            "partition_description": "TEXT",
            "table_rows": "INT",
            "avg_row_length": "INT",
            "data_length": "INT",
            "max_data_length": "INT",
            "index_length": "INT",
            "data_free": "INT",
            "create_time": "TEXT",
            "update_time": "TEXT",
            "check_time": "TEXT",
            "checksum": "TEXT",
            "partition_comment": "TEXT",
            "nodegroup": "TEXT",
            "tablespace_name": "TEXT",
        },
        "referential_constraints": {
            "constraint_catalog": "TEXT",
            "constraint_schema": "TEXT",
            "constraint_name": "TEXT",
            "unique_constraint_catalog": "TEXT",
            "unique_constraint_schema": "TEXT",
            "unique_constraint_name": "TEXT",
            "match_option": "TEXT",
            "update_rule": "TEXT",
            "delete_rule": "TEXT",
            "table_name": "TEXT",
            "referenced_table_name": "TEXT",
        },
        "routines": {
            "specific_name": "TEXT",
            "routine_catalog": "TEXT",
            "routine_schema": "TEXT",
            "routine_name": "TEXT",
            "routine_type": "TEXT",
            "data_type": "TEXT",
            "character_maximum_length": "INT",
            "character_octet_length": "INT",
            "numeric_precision": "INT",
            "numeric_scale": "INT",
            "datetime_precision": "INT",
            "character_set_name": "TEXT",
            "collation_name": "TEXT",
            "dtd_identifier": "TEXT",
            "routine_body": "TEXT",
            "routine_definition": "TEXT",
            "external_name": "TEXT",
            "external_language": "TEXT",
            "parameter_style": "TEXT",
            "is_deterministic": "TEXT",
            "sql_data_access": "TEXT",
            "sql_path": "TEXT",
            "security_type": "TEXT",
            "created": "TEXT",
            "last_altered": "TEXT",
            "sql_mode": "TEXT",
            "routine_comment": "TEXT",
            "definer": "TEXT",
            "character_set_client": "TEXT",
            "collation_connection": "TEXT",
            "database_collation": "TEXT",
        },
        "schema_privileges": {
            "grantee": "TEXT",
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "privilege_type": "TEXT",
            "is_grantable": "TEXT",
        },
        "schemata": {
            "catalog_name": "TEXT",
            "schema_name": "TEXT",
            "default_character_set_name": "TEXT",
            "default_collation_name": "TEXT",
            "sql_path": "TEXT",
        },
        "statistics": {
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "non_unique": "INT",
            "index_schema": "TEXT",
            "index_name": "TEXT",
            "seq_in_index": "INT",
            "column_name": "TEXT",
            "collation": "TEXT",
            "cardinality": "INT",
            "sub_part": "TEXT",
            "packed": "TEXT",
            "nullable": "TEXT",
            "index_type": "TEXT",
            "comment": "TEXT",
            "index_comment": "TEXT",
            "is_visible": "TEXT",
            "expression": "TEXT",
        },
        "tables": {
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "table_type": "TEXT",
            "engine": "TEXT",
            "version": "TEXT",
            "row_format": "TEXT",
            "table_rows": "INT",
            "avg_row_length": "INT",
            "data_length": "INT",
            "max_data_length": "INT",
            "index_length": "INT",
            "data_free": "INT",
            "auto_increment": "INT",
            "create_time": "TEXT",
            "update_time": "TEXT",
            "check_time": "TEXT",
            "table_collation": "TEXT",
            "checksum": "TEXT",
            "create_options": "TEXT",
            "table_comment": "TEXT",
        },
        "table_constraints": {
            "constraint_name": "TEXT",
            "constraint_catalog": "TEXT",
            "constraint_schema": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "constraint_type": "TEXT",
            "enforced": "TEXT",
        },
        "table_privileges": {
            "grantee": "TEXT",
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "privilege_type": "TEXT",
            "is_grantable": "TEXT",
        },
        "triggers": {
            "trigger_catalog": "TEXT",
            "trigger_schema": "TEXT",
            "trigger_name": "TEXT",
            "event_manipulation": "TEXT",
            "event_object_catalog": "TEXT",
            "event_object_schema": "TEXT",
            "event_object_table": "TEXT",
            "action_order": "INT",
            "action_condition": "TEXT",
            "action_statement": "TEXT",
            "action_orientation": "TEXT",
            "action_timing": "TEXT",
            "action_reference_old_table": "TEXT",
            "action_reference_new_table": "TEXT",
            "action_reference_old_row": "TEXT",
            "action_reference_new_row": "TEXT",
            "created": "TEXT",
            "sql_mode": "TEXT",
            "definer": "TEXT",
            "character_set_client": "TEXT",
            "collation_connection": "TEXT",
            "database_collation": "TEXT",
        },
        "user_privileges": {
            "grantee": "TEXT",
            "table_catalog": "TEXT",
            "privilege_type": "TEXT",
            "is_grantable": "TEXT",
        },
        "views": {
            "table_catalog": "TEXT",
            "table_schema": "TEXT",
            "table_name": "TEXT",
            "view_definition": "TEXT",
            "check_option": "TEXT",
            "is_updatable": "TEXT",
            "definer": "TEXT",
            "security_type": "TEXT",
            "character_set_client": "TEXT",
            "collation_connection": "TEXT",
        },
    },
    "mysql": {
        "procs_priv": {
            "host": "TEXT",
            "db": "TEXT",
            "user": "TEXT",
            "routine_name": "TEXT",
            "routine_type": "TEXT",
            "proc_priv": "TEXT",
            "timestamp": "TEXT",
            "grantor": "TEXT",
        },
        "role_edges": {
            "from_host": "TEXT",
            "from_user": "TEXT",
            "to_host": "TEXT",
            "to_user": "TEXT",
            "with_admin_option": "TEXT",
        },
        "user": {
            "host": "TEXT",
            "user": "TEXT",
            "select_priv": "TEXT",
            "insert_priv": "TEXT",
            "update_priv": "TEXT",
            "delete_priv": "TEXT",
            "index_priv": "TEXT",
            "alter_priv": "TEXT",
            "create_priv": "TEXT",
            "drop_priv": "TEXT",
            "grant_priv": "TEXT",
            "create_view_priv": "TEXT",
            "show_view_priv": "TEXT",
            "create_routine_priv": "TEXT",
            "alter_routine_priv": "TEXT",
            "execute_priv": "TEXT",
            "trigger_priv": "TEXT",
            "event_priv": "TEXT",
            "create_tmp_table_priv": "TEXT",
            "lock_tables_priv": "TEXT",
            "references_priv": "TEXT",
            "reload_priv": "TEXT",
            "shutdown_priv": "TEXT",
            "process_priv": "TEXT",
            "file_priv": "TEXT",
            "show_db_priv": "TEXT",
            "super_priv": "TEXT",
            "repl_slave_priv": "TEXT",
            "repl_client_priv": "TEXT",
            "create_user_priv": "TEXT",
            "create_tablespace_priv": "TEXT",
            "create_role_priv": "TEXT",
            "drop_role_priv": "TEXT",
            "ssl_type": "TEXT",
            "ssl_cipher": "TEXT",
            "x509_issuer": "TEXT",
            "x509_subject": "TEXT",
            "plugin": "TEXT",
            "authentication_string": "TEXT",
            "password_expired": "TEXT",
            "password_last_changed": "TEXT",
            "password_lifetime": "TEXT",
            "account_locked": "TEXT",
            "password_reuse_history": "TEXT",
            "password_reuse_time": "TEXT",
            "password_require_current": "TEXT",
            "user_attributes": "TEXT",
            "max_questions": "INT",
            "max_updates": "INT",
            "max_connections": "INT",
            "max_user_connections": "INT",
        },
    },
}



from typing import Any, Mapping, Callable
from datetime import datetime

Functions = Mapping[str, Callable[[], Any]]


def mysql_datetime_function_mapping(timestamp: datetime) -> Functions:
    functions = {
        "NOW": lambda: timestamp.strftime("%Y-%m-%d %H:%M:%S"),
        "CURDATE": lambda: timestamp.strftime("%Y-%m-%d"),
        "CURTIME": lambda: timestamp.strftime("%H:%M:%S"),
    }
    functions.update(
        {
            "CURRENT_TIMESTAMP": functions["NOW"],
            "LOCALTIME": functions["NOW"],
            "LOCALTIMESTAMP": functions["NOW"],
            "CURRENT_DATE": functions["CURDATE"],
            "CURRENT_TIME": functions["CURTIME"],
        }
    )
    return functions



from __future__ import annotations

import random
from typing import Dict, Optional, TYPE_CHECKING

from mysql_mimic.constants import KillKind
from mysql_mimic.utils import seq

if TYPE_CHECKING:
    from mysql_mimic.connection import Connection


class TooManyConnections(Exception):
    pass


class Control:
    """
    Base class for controlling server state.

    This is intended to encapsulate operations that might span deployment of multiple
    MySQL mimic server instances.
    """

    async def add(self, connection: Connection) -> int:
        """
        Add a new connection

        Returns:
            connection id
        """
        raise NotImplementedError()

    async def remove(self, connection_id: int) -> None:
        """Remove an existing connection"""
        raise NotImplementedError()

    async def kill(
        self, connection_id: int, kind: KillKind = KillKind.CONNECTION
    ) -> None:
        """Request termination of an existing connection"""
        raise NotImplementedError()


class LocalControl(Control):
    """
    Simple Control implementation that handles everything in-memory.

    Args:
        server_id: set a unique server ID. This is used to generate globally unique
            connection IDs. This should be an integer between 0 and 65535.
            If left as None, a random server ID will be generated.
    """

    _CONNECTION_ID_BITS = 16
    _MAX_CONNECTION_SEQ = 2**_CONNECTION_ID_BITS
    _MAX_SERVER_ID = 2**16

    def __init__(self, server_id: Optional[int] = None):
        self._connection_seq = seq(self._MAX_CONNECTION_SEQ)
        self._connections: Dict[int, Connection] = {}
        self.server_id = server_id or random.randint(0, self._MAX_SERVER_ID - 1)

    async def add(self, connection: Connection) -> int:
        connection_id = self._new_connection_id()
        self._connections[connection_id] = connection
        return connection_id

    async def remove(self, connection_id: int) -> None:
        self._connections.pop(connection_id, None)

    async def kill(
        self, connection_id: int, kind: KillKind = KillKind.CONNECTION
    ) -> None:
        conn = self._connections.get(connection_id)
        if conn:
            conn.kill(kind)

    def _new_connection_id(self) -> int:
        """
        Generate a connection ID.

        MySQL connection IDs are 4 bytes.

        This is tricky for us, as there may be multiple MySQL-Mimic server instances.

        We use a concatenation of the server ID (first two bytes) and connection
        sequence (second two bytes):

        |<---server ID--->|<-conn sequence->|
         00000000 00000000 00000000 00000000

        If incremental connection IDs aren't manually provided, this isn't guaranteed
        to be unique. But collisions should be highly unlikely.
        """
        if len(self._connections) >= self._MAX_CONNECTION_SEQ:
            raise TooManyConnections()

        server_id_prefix = (
            self.server_id % self._MAX_SERVER_ID
        ) << self._CONNECTION_ID_BITS

        connection_id = server_id_prefix + next(self._connection_seq)

        # Avoid connection ID collisions in the unlikely chance that a connection is
        # alive longer than it takes for the sequence to reset.
        while connection_id in self._connections:
            connection_id = server_id_prefix + next(self._connection_seq)

        return connection_id



from contextlib import contextmanager
from typing import Dict, Mapping, Generator

from sqlglot import expressions as exp

from mysql_mimic.intercept import value_to_expression, expression_to_value
from mysql_mimic.variables import Variables

variable_constants = {
    "CURRENT_USER",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_DATE",
}


def _get_var_assignments(expression: exp.Expression) -> Dict[str, str]:
    """Returns a dictionary of session variables to replace, as indicated by SET_VAR hints."""
    hints = expression.find_all(exp.Hint)
    if not hints:
        return {}

    assignments = {}

    # Iterate in reverse order so higher SET_VAR hints get priority
    for hint in reversed(list(hints)):
        set_var_hint = None

        for e in hint.expressions:
            if isinstance(e, exp.Func) and e.name == "SET_VAR":
                set_var_hint = e
                for eq in e.expressions:
                    assignments[eq.left.name] = expression_to_value(eq.right)

        if set_var_hint:
            set_var_hint.pop()

        # Remove the hint entirely if SET_VAR was the only expression
        if not hint.expressions:
            hint.pop()

    return assignments


class VariableProcessor:
    """
    This class modifies the query in two ways:
        1. Processing SET_VAR hints for system variables in the query text
        2. Replacing functions in the query with their replacements defined in the functions argument.
    original values.
    """

    def __init__(
        self, functions: Mapping, variables: Variables, expression: exp.Expression
    ):
        self._functions = functions
        self._variables = variables
        self._expression = expression

        # Stores the original system variable values.
        self._orig: Dict[str, str] = {}

    @contextmanager
    def set_variables(self) -> Generator[exp.Expression, None, None]:
        assignments = _get_var_assignments(self._expression)
        self._orig = {k: self._variables.get(k) for k in assignments}
        for k, v in assignments.items():
            self._variables.set(k, v)

        self._replace_variables()

        yield self._expression

        for k, v in self._orig.items():
            self._variables.set(k, v)

    def _replace_variables(self) -> None:
        """Replaces certain functions in the query with literals provided from the mapping in _functions,
        and session parameters with the values of the session variables.
        """
        if isinstance(self._expression, exp.Set):
            for setitem in self._expression.expressions:
                if isinstance(setitem.this, exp.Binary):
                    # In the case of statements like: SET @@foo = @@bar
                    # We only want to replace variables on the right
                    setitem.this.set(
                        "expression",
                        setitem.this.expression.transform(self._transform, copy=True),
                    )
        else:
            self._expression.transform(self._transform, copy=False)

    def _transform(self, node: exp.Expression) -> exp.Expression:
        new_node = None

        if isinstance(node, exp.Func):
            if isinstance(node, exp.Anonymous):
                func_name = node.name.upper()
            else:
                func_name = node.sql_name()
            func = self._functions.get(func_name)
            if func:
                value = func()
                new_node = value_to_expression(value)
        elif isinstance(node, exp.Column) and node.sql() in variable_constants:
            value = self._functions[node.sql()]()
            new_node = value_to_expression(value)
        elif isinstance(node, exp.SessionParameter):
            value = self._variables.get(node.name)
            new_node = value_to_expression(value)

        if (
            new_node
            and isinstance(node.parent, exp.Select)
            and node.arg_key == "expressions"
        ):
            new_node = exp.alias_(new_node, exp.to_identifier(node.sql()))

        return new_node or node



from __future__ import annotations

import re
from collections import defaultdict
from itertools import chain
from dataclasses import dataclass
from typing import Any, Optional, List, Dict, Iterable

from sqlglot.executor import Table, execute
from sqlglot import expressions as exp

from mysql_mimic.constants import INFO_SCHEMA
from mysql_mimic.results import AllowedResult
from mysql_mimic.errors import MysqlError, ErrorCode
from mysql_mimic.packets import ComFieldList
from mysql_mimic.utils import dict_depth


@dataclass
class Column:
    name: str
    type: str
    table: str
    is_nullable: bool = True
    default: Optional[str] = None
    comment: Optional[str] = None
    schema: Optional[str] = None
    catalog: Optional[str] = "def"


def mapping_to_columns(schema: dict) -> List[Column]:
    """Convert a schema mapping into a list of Column instances.

    Example schema defining columns by type:
     {
         "customer_table" : {
           "first_name" : "TEXT"
           "last_name" : "TEXT"
           "id" : "INT"
         },
         "sales_table" : {
           "ds" : "DATE"
           "customer_id" : "INT"
           "amount" : "DOUBLE"
         }
    }
    """
    depth = dict_depth(schema)
    if depth < 2:
        return []
    if depth == 2:
        # {table: {col: type}}
        schema = {"": schema}
        depth += 1
    if depth == 3:
        # {db: {table: {col: type}}}
        schema = {"def": schema}  # def is the default MySQL catalog
        depth += 1
    if depth != 4:
        raise MysqlError("Invalid schema mapping")

    result = []
    for catalog, dbs in schema.items():
        for db, tables in dbs.items():
            for table, cols in tables.items():
                for column, coltype in cols.items():
                    result.append(
                        Column(
                            name=column,
                            type=coltype,
                            table=table,
                            schema=db,
                            catalog=catalog,
                        )
                    )

    return result


def info_schema_tables(columns: Iterable[Column]) -> Dict[str, Dict[str, Table]]:
    """
    Convert a list of Column instances into a mapping of SQLGlot Tables.

    These Tables are used by SQLGlot to execute INFORMATION_SCHEMA queries.
    """
    ordinal_positions: dict[Any, int] = defaultdict(lambda: 0)

    data = {
        db: {k: Table(tuple(v.keys())) for k, v in tables.items()}
        for db, tables in INFO_SCHEMA.items()
    }

    tables = set()
    dbs = set()
    catalogs = set()

    info_schema_cols = mapping_to_columns(INFO_SCHEMA)

    for column in chain(columns, info_schema_cols):
        tables.add((column.catalog, column.schema, column.table))
        dbs.add((column.catalog, column.schema))
        catalogs.add(column.catalog)
        key = (column.catalog, column.schema, column.table)
        ordinal_position = ordinal_positions[key]
        ordinal_positions[key] += 1
        data["information_schema"]["columns"].append(
            (
                column.catalog,  # table_catalog
                column.schema,  # table_schema
                column.table,  # table_name
                column.name,  # column_name
                ordinal_position,  # ordinal_position
                column.default,  # column_default
                "YES" if column.is_nullable else "NO",  # is_nullable
                column.type,  # data_type
                None,  # character_maximum_length
                None,  # character_octet_length
                None,  # numeric_precision
                None,  # numeric_scale
                None,  # datetime_precision
                "NULL",  # character_set_name
                "NULL",  # collation_name
                column.type,  # column_type
                None,  # column_key
                None,  # extra
                None,  # privileges
                column.comment,  # column_comment
                None,  # generation_expression
                None,  # srs_id
            )
        )

    for catalog, db, table in sorted(tables):
        data["information_schema"]["tables"].append(
            (
                catalog,  # table_catalog
                db,  # table_schema
                table,  # table_name
                "SYSTEM TABLE" if db in INFO_SCHEMA else "BASE TABLE",  # table_type
                "MinervaSQL",  # engine
                "1.0",  # version
                None,  # row_format
                None,  # table_rows
                None,  # avg_row_length
                None,  # data_length
                None,  # max_data_length
                None,  # index_length
                None,  # data_free
                None,  # auto_increment
                None,  # create_time
                None,  # update_time
                None,  # check_time
                "utf8mb4_general_ci ",  # table_collation
                None,  # checksum
                None,  # create_options
                None,  # table_comment
            )
        )

    for catalog, db in sorted(dbs):
        data["information_schema"]["schemata"].append(
            (
                catalog,  # catalog_name
                db,  # schema_name
                "utf8mb4",  # default_character_set_name
                "utf8mb4_general_ci",  # default_collation_name
                None,  # sql_path
            )
        )

    return data


def show_statement_to_info_schema_query(
    show: exp.Show, database: Optional[str] = None
) -> exp.Select:
    kind = show.name.upper()
    if kind == "COLUMNS":
        outputs = [
            'column_name AS "Field"',
            'data_type AS "Type"',
            'is_nullable AS "Null"',
            'column_key AS "Key"',
            'column_default AS "Default"',
            'extra AS "Extra"',
        ]
        if show.args.get("full"):
            outputs.extend(
                [
                    'collation_name AS "Collation"',
                    'privileges AS "Privileges"',
                    'column_comment AS "Comment"',
                ]
            )
        table = show.text("target")
        if not table:
            raise MysqlError(
                "You have an error in your SQL syntax. Table name is missing.",
                code=ErrorCode.PARSE_ERROR,
            )
        select = (
            exp.select(*outputs)
            .from_("information_schema.columns")
            .where(f"table_name = '{table}'")
        )
        db = show.text("db") or database
        if db:
            select = select.where(f"table_schema = '{db}'")
        like = show.text("like")
        if like:
            select = select.where(f"column_name LIKE '{like}'")
    elif kind == "TABLES":
        outputs = ['table_name AS "Table_name"']
        if show.args.get("full"):
            outputs.extend(['table_type AS "Table_type"'])

        select = exp.select(*outputs).from_("information_schema.tables")
        db = show.text("db") or database
        if not db:
            raise MysqlError("No database selected.", code=ErrorCode.NO_DB_ERROR)
        select = select.where(f"table_schema = '{db}'")
        like = show.text("like")
        if like:
            select = select.where(f"table_name LIKE '{like}'")
    elif kind == "DATABASES":
        select = exp.select('schema_name AS "Database"').from_(
            "information_schema.schemata"
        )
        like = show.text("like")
        if like:
            select = select.where(f"schema_name LIKE '{like}'")
    elif kind == "INDEX":
        outputs = [
            '"table_name" AS Table',
            '"non_unique" AS Non_unique',
            '"index_name" AS Key_name',
            '"seq_in_index" AS Seq_in_index',
            '"column_name" AS Column_name',
            '"collation" AS Collation',
            '"cardinality" AS Cardinality',
            '"sub_part" AS Sub_part',
            '"packed" AS Packed',
            '"nullable" AS Null',
            '"index_type" AS Index_type',
            '"comment" AS Comment',
            '"index_comment" AS Index_comment',
            '"is_visible" AS Visible',
            '"expression" AS Expression',
        ]
        table = show.text("target")
        if not table:
            raise MysqlError(
                "You have an error in your SQL syntax. Table name is missing.",
                code=ErrorCode.PARSE_ERROR,
            )
        select = (
            exp.select(*outputs)
            .from_("information_schema.statistics")
            .where(f"table_name = '{table}'")
        )
        db = show.text("db") or database
        if db:
            select = select.where(f"table_schema = '{db}'")
    else:
        raise MysqlError(
            f"Unsupported SHOW command: {kind}", code=ErrorCode.NOT_SUPPORTED_YET
        )

    return select


def com_field_list_to_show_statement(com_field_list: ComFieldList) -> str:
    show = exp.Show(
        this="COLUMNS",
        target=exp.to_identifier(com_field_list.table),
    )
    if com_field_list.wildcard:
        show.set("like", exp.Literal.string(com_field_list.wildcard))
    return show.sql(dialect="mysql")


def like_to_regex(like: str) -> re.Pattern:
    like = like.replace("%", ".*?")
    like = like.replace("_", ".")
    return re.compile(like)


class BaseInfoSchema:
    """
    Base InfoSchema interface used by the `Session` class.
    """

    async def query(self, expression: exp.Expression) -> AllowedResult: ...


class InfoSchema(BaseInfoSchema):
    """
    InfoSchema implementation that uses SQLGlot to execute queries.
    """

    def __init__(self, tables: Dict[str, Dict[str, Table]]):
        self.tables = tables

    async def query(self, expression: exp.Expression) -> AllowedResult:
        expression = self._preprocess(expression)
        result = execute(expression, schema=INFO_SCHEMA, tables=self.tables)
        return result.rows, result.columns

    @classmethod
    def from_mapping(cls, mapping: dict) -> InfoSchema:
        columns = mapping_to_columns(mapping)
        return cls(info_schema_tables(columns))

    @classmethod
    def from_columns(cls, columns: List[Column]) -> InfoSchema:
        return cls(info_schema_tables(columns))

    def _preprocess(self, expression: exp.Expression) -> exp.Expression:
        return expression.transform(_remove_collate)


def ensure_info_schema(schema: dict | BaseInfoSchema) -> BaseInfoSchema:
    if isinstance(schema, BaseInfoSchema):
        return schema
    return InfoSchema.from_mapping(schema)


def _remove_collate(node: exp.Expression) -> exp.Expression:
    """
    SQLGlot's executor doesn't support collation specifiers.

    Just remove them for now.
    """
    if isinstance(node, exp.Collate):
        return node.left
    return node



import pytest
from sqlglot import expressions as exp

from mysql_mimic.results import ensure_result_set
from mysql_mimic.schema import (
    Column,
    InfoSchema,
    mapping_to_columns,
    show_statement_to_info_schema_query,
)
from mysql_mimic.utils import aiterate


@pytest.mark.asyncio
def test_mapping_to_columns() -> None:
    schema = {
        "table_1": {
            "col_1": "TEXT",
            "col_2": "INT",
        }
    }

    columns = mapping_to_columns(schema=schema)

    assert columns[0] == Column(
        name="col_1", type="TEXT", table="table_1", schema="", catalog="def"
    )
    assert columns[1] == Column(
        name="col_2", type="INT", table="table_1", schema="", catalog="def"
    )


@pytest.mark.asyncio
async def test_info_schema_from_columns() -> None:
    input_columns = [
        Column(
            name="col_1",
            type="TEXT",
            table="table_1",
            schema="my_db",
            catalog="def",
            comment="This is a comment",
        ),
        Column(
            name="col_1", type="TEXT", table="table_2", schema="my_db", catalog="def"
        ),
    ]
    schema = InfoSchema.from_columns(columns=input_columns)
    table_query = show_statement_to_info_schema_query(exp.Show(this="TABLES"), "my_db")
    result_set = await ensure_result_set(await schema.query(table_query))
    table_names = [row[0] async for row in aiterate(result_set.rows)]
    assert table_names == ["table_1", "table_2"]

    column_query = show_statement_to_info_schema_query(
        exp.Show(this="COLUMNS", full=True, target="table_1"), "my_db"
    )
    column_results = await ensure_result_set(await schema.query(column_query))
    columns = [row async for row in aiterate(column_results.rows)]
    assert columns == [
        (
            "col_1",
            "TEXT",
            "YES",
            None,
            None,
            None,
            "NULL",
            None,
            "This is a comment",
        )
    ]
